/*
	Sprankle Character Map - Inline, dynamic character map.
    Copyright (C) 2004  Steven Wittens

	see LICENSE.TXT for license.
*/
#include "sprankle.h"

BOOL CALLBACK MainDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void SprankleOn(HWND h);
void SprankleSend(wstring s);
void SprankleOff(HWND h);
void SprankleBase(wchar_t c);
void SprankleTitle();
void FitMap();
void SprankleGrid(SprankleTable *current, DRAWITEMSTRUCT *dis);

/* Global vars */
HWND lastfocus;
HWND control;
HWND dlg;
SprankleTable *current;
HFONT font1;
HFONT font2;
HFONT font3;
int sel;

/* One square for a character.
Could be expanded for stuff like sequences for entering "e'" = "".*/
struct MapSquare {
	wchar_t c;
};

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	MSG msg;

	sel = 0;
	dlg = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, MainDialogProc);

	font1 = CreateFont(
		34, 0, 0, 0, FW_NORMAL,
		false, false, false,
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
		DEFAULT_PITCH|FF_SWISS, L"Arial Unicode MS");
	font2 = CreateFont(
		12, 0, 0, 0, FW_NORMAL,
		false, false, false,
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
		DEFAULT_PITCH|FF_SWISS, L"Tahoma");
	font3 = CreateFont(
		12, 0, 0, 0, FW_NORMAL,
		false, false, false,
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
		DEFAULT_PITCH|FF_SWISS, L"Marlett");

	SprankleTable::InitTable();

	while (GetMessage(&msg, NULL, 0, 0)) {
		if (msg.message == WM_CHAR) {
			// Set base character
			SprankleBase(msg.wParam);
		}
		if (msg.message == WM_KEYDOWN) {
			if ((msg.wParam == VK_ESCAPE) || (msg.wParam == VK_BACK)) {
				// Hide sprankle
				SprankleOff(control);
			}
			if (current) {
				if (msg.wParam == VK_LEFT) {
					// Move selectbox left
					sel = sel - 1;
					if (sel < 0) {
						sel = current->dest.size() - 1;
					}
				}
				else if (msg.wParam == VK_RIGHT) {
					// Move selectbox right
					sel = sel + 1;
					if (sel >= current->dest.size()) {
						sel = 0;
					}
				}
				InvalidateRect(GetDlgItem(dlg, IDC_MAP), 0, false);
				SprankleTitle();
			}
		}
		if (!IsDialogMessage(dlg, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	};

	DeleteObject(font2);
	DeleteObject(font1);

	return 0;
};

// Show sprankle map on h
void SprankleOn(HWND h) {
	ShowWindow(dlg, SW_SHOW);
	SetForegroundWindow(dlg);
	SetFocus(dlg);
	RECT rh, rd;
	GetWindowRect(h, &rh);
	GetWindowRect(dlg, &rd);
	SetWindowPos(dlg, 0, rh.left, rh.top - rd.bottom + rd.top, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
};

// Send string s to active control
void SprankleSend(wstring s) {
	for (int i = 0; i < s.length(); ++i) {
		if (s[i] != 0) {
			SendMessage(control, WM_CHAR, s[i], 0);
		}
	}
};

// Hide sprankle map
void SprankleOff(HWND h) {
	ShowWindow(dlg, SW_HIDE);
	SetForegroundWindow(control);
	SetFocus(control);
	control = 0;
};

// Update titlebar
void SprankleTitle() {
	// Update title bar
	wchar_t title[256];
	if (current->dest.size() > 10) {
		wsprintf(title, L"Sprankle - %d/%d", sel/10 + 1, (current->dest.size() - 1)/10 + 1);
	}
	else {
		wsprintf(title, L"Sprankle");
	}
	SetWindowText(dlg, title);
}

// Set base character to c
void SprankleBase(wchar_t c) {
	int page = (sel/10*10);
	// Key handling
	if (c == ' ') {
		SprankleSend(current->dest[sel]);
		SprankleOff(control);
	}
	if (c >= '0' && c <= '9' && current) {
		if (c == '0') {
			c = '9' + 1;
		}
		if ((c - '1' + page) < current->dest.size()) {
			SprankleSend(current->dest[c - '1' + page]);
		}
		SprankleOff(control);
	}

	// Table lookup
	vector<SprankleTable>::iterator i;
	for (i = table.begin(); i != table.end(); ++i) {
		if (i->source == c) {
			break;
		}
	}
	if (i != table.end()) {
		if ((i == current) && (current)) {
			// Same base character, skip to next page
			sel = sel + 10;
			// Out of range
			if (sel >= current->dest.size()) {
				// If we skipped beyond the last page, go back to the beginning
				if (sel/10*10 >= current->dest.size()) {
					sel = sel % 10;
				}
				// Else we just select the last character on this (short) page
				else {
					sel = current->dest.size() - 1;
				}
			}
		}
		else {
			// New base character
			current = i;
			// Validate selection
			sel = min(sel, current->dest.size() - 1);
		}
		InvalidateRect(GetDlgItem(dlg, IDC_MAP), 0, false);
		SprankleTitle();
	}
};

// Fit map to dialog size
void FitMap() {
	HWND m = GetDlgItem(dlg, IDC_MAP);
	RECT r;
	GetClientRect(dlg, &r);
	SetWindowPos(m, 0, 0, 0, r.right, r.bottom, SWP_NOZORDER);
};

#define GRID 30
// Draw grid of characters
void SprankleGrid(SprankleTable *current, DRAWITEMSTRUCT *dis) {
	// First index on current page
	int page = sel/10*10;

	// Calculate width of current page
	int w;
	if (current) {
		w = 0;
		for (int i = page; i < min(page + 10, current->dest.size()); ++i) {
			int l = current->dest[i].length();
			wstring *str = &current->dest[i];
			w += (l + 1) * GRID / 2;
		}
	}
	else {
		w = GRID;
	}
	RECT r1, r2;
	GetWindowRect(dlg, &r1);
	GetWindowRect(dlg, &r2);

	int bottom = r2.bottom;

	// Perform opposite adjustment
	int style = GetWindowLong(dlg, GWL_STYLE);
	int exstyle = GetWindowLong(dlg, GWL_EXSTYLE);
	AdjustWindowRectEx(&r1, style, false, exstyle);

	// Invert adjustment
	r2.left += r2.left - r1.left;

	// Set gridsize
	r2.top = 0;
	r2.right = r2.left + w;
	r2.bottom = r2.top + GRID;

	// Adjust to full size
	AdjustWindowRectEx(&r2, style, false, exstyle);

	r2.top = bottom - (r2.bottom - r2.top);
	r2.bottom = bottom;

	// Change size
	SetWindowPos(dlg, 0, r2.left, r2.top, r2.right - r2.left, r2.bottom - r2.top, SWP_NOZORDER);

	// Fit map
	FitMap();

	// Fill BG
	RECT r;
	GetClientRect(dis->hwndItem, &r);
	FillRect(dis->hDC, &r, (HBRUSH)GetStockObject(WHITE_BRUSH));

	// Draw letters
	if (current) {
		vector<wstring>::iterator i;

		r1.top = 0;
		r1.left = 0;
		r1.right = GRID;
		r1.bottom = GRID;

		SetBkMode(dis->hDC, TRANSPARENT);
		
		HBRUSH br = CreateSolidBrush(0x7DEEFF);
		HFONT old;
		HDC dc = GetDC(0);
		old = (HFONT)SelectObject(dc, font1);
		SelectObject(dc, old);
		old = (HFONT)SelectObject(dc, font2);
		SelectObject(dc, old);
		ReleaseDC(0, dc);

		HPEN pen, oldp;
		pen = CreatePen(PS_SOLID, 1, 0xC0C0C0);
		oldp = (HPEN)SelectObject(dis->hDC, pen);

		old = (HFONT)SelectObject(dis->hDC, font1);
		
		int j = 0;
		i = &current->dest[page];
		for (; i != current->dest.end() && j < 10; ++i) {
			wchar_t str[2] = { 0, 0 };

			r1.right = r1.left + (i->length() + 1) * GRID / 2;
			r1.left++;

			// Selected character
			if (sel - page == j) {
				FillRect(dis->hDC, &r1, br);
			}

			r1.left--;

			MoveToEx(dis->hDC, r1.right, r1.top, 0);
			LineTo(dis->hDC, r1.right, r1.bottom);

			// Draw left/right arrows
			SelectObject(dis->hDC, font3);
			SetTextColor(dis->hDC, 0x0000A0);
			if ((j == 0) && page) {
				r1.left--;
				DrawText(dis->hDC, L"3", 1, &r1, DT_SINGLELINE|DT_LEFT|DT_TOP);
				r1.left++;
			}
			if ((j == 9) && (page + 10 < current->dest.size())) {
				r1.right+=2;
				DrawText(dis->hDC, L"4", 1, &r1, DT_SINGLELINE|DT_RIGHT|DT_TOP);
				r1.right-=2;
			}

			// Draw hint numbers
			SelectObject(dis->hDC, font2);
			SetTextColor(dis->hDC, 0x404040);
			str[0] = '1' + ((++j)%10) - 1;
			r1.left += 3;
			DrawText(dis->hDC, str, 1, &r1, DT_SINGLELINE|DT_RIGHT|DT_BOTTOM);
			r1.left -= 3;

			// Draw character
			SelectObject(dis->hDC, font1);
			SetTextColor(dis->hDC, 0x0);
			DrawText(dis->hDC, i->c_str(), -1, &r1, DT_SINGLELINE|DT_CENTER|DT_VCENTER);
			r1.left = r1.right;
		}

		DeleteObject(br);

		SelectObject(dis->hDC, oldp);
		SelectObject(dis->hDC, old);
	}
};

BOOL CALLBACK MainDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_TIMER:
		{
			HWND h = GetForegroundWindow();
			if ((h != hwndDlg) && (h != 0)) {
				lastfocus = h;
			}
		}
		return true;

	case WM_DRAWITEM:
		{
			int idCtl = wParam;
			DRAWITEMSTRUCT *dis = (DRAWITEMSTRUCT *)lParam;

			if (idCtl == IDC_MAP) {
				SprankleGrid(current, dis);
			}
		}
		return true;

	case WM_INITDIALOG:
		{
			RegisterHotKey(hwndDlg, 0x1, MOD_WIN, 0x53);
			RegisterHotKey(hwndDlg, 0x2, MOD_WIN|MOD_SHIFT, 0x53);
			RegisterHotKey(hwndDlg, 0x3, MOD_CONTROL|MOD_ALT, 0x53);
			RegisterHotKey(hwndDlg, 0x4, MOD_CONTROL|MOD_ALT|MOD_SHIFT, 0x53);

			SetTimer(hwndDlg, 1, 50, 0);

			FitMap();
			SetWindowLong(GetDlgItem(hwndDlg, IDC_MAP), GWL_STYLE, SS_OWNERDRAW | GetWindowLong(GetDlgItem(hwndDlg, IDC_MAP), GWL_STYLE));

//			ShowWindow(hwndDlg, SW_SHOW);
		}
		return true;

	case WM_HOTKEY:
		if (LOWORD(lParam) & MOD_SHIFT) {
			DestroyWindow(hwndDlg);
		}
		else if ((lastfocus != hwndDlg) && (!control)) {
			unsigned long process;
			int res = GetWindowThreadProcessId(lastfocus, &process);
			if (AttachThreadInput(GetCurrentThreadId(), res, true)) {
				control = GetFocus();
				AttachThreadInput(GetCurrentThreadId(), res, false);
				SprankleOn(control);
			}
		}
		return true;

	case WM_ACTIVATE:
		if (LOWORD(wParam) == WA_INACTIVE) {
			control = 0;
		}
		return true;

	case WM_CLOSE:
		DestroyWindow(hwndDlg);
		return true;

	case WM_DESTROY:
		KillTimer(hwndDlg, 1);
		UnregisterHotKey(hwndDlg, 0x1);
		UnregisterHotKey(hwndDlg, 0x2);
		UnregisterHotKey(hwndDlg, 0x3);
		UnregisterHotKey(hwndDlg, 0x4);
		PostQuitMessage(0);
		return true;
	}
	return false;
};
